/*
 * Copyright 2009, Josh Devins, Jitr.org.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.jitr;

import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.net.ServerSocket;
import java.util.Collection;
import java.util.LinkedList;

import org.apache.commons.lang.Validate;
import org.jitr.core.JitrException;

/**
 * Provides various simple utility methods.
 * 
 * @author Josh Devins (joshdevins@jitr.org)
 */
public final class JitrUtils {

    private JitrUtils() {
        // non-instantiable class
    }

    /**
     * Finds any declared fields in the given class with the given annotation.
     */
    public static Collection<Field> getAnnotatedDeclaredFields(final Class<?> clazz,
            final Class<? extends Annotation> annotationClass) {

        validateClassesNotNull(clazz, annotationClass);

        final Field[] fields = clazz.getDeclaredFields();
        final Collection<Field> foundFields = new LinkedList<Field>();

        for (Field field : fields) {

            if (field.getAnnotation(annotationClass) != null) {
                foundFields.add(field);
            }
        }

        return foundFields;
    }

    /**
     * Finds an annotated field amongst a hierarchy of classes. Should multiple classes in the
     * hierarchy have a field with the same annotation, the field from the most specific or lowest
     * subclass is used.
     * 
     * @throws JitrException
     *             When multiple annotated fields are found in any one class.
     */
    public static Field getMostSpecificAnnotatedDeclaredField(final Class<?> clazz,
            final Class<? extends Annotation> annotationClass) {

        validateClassesNotNull(clazz, annotationClass);

        // if this is the top of the hierarchy, end here
        if (Object.class.equals(clazz)) {
            return null;
        }

        // first look in this class
        final Collection<Field> fields = getAnnotatedDeclaredFields(clazz, annotationClass);
        if (fields.size() > 1) {
            throw new JitrException("One field was expected, but multiple fields in "
                    + clazz.getSimpleName() + " were annotated with "
                    + annotationClass.getSimpleName());
        }

        // nothing found, recurse and try the next class
        if (fields.isEmpty()) {
            return getMostSpecificAnnotatedDeclaredField(clazz.getSuperclass(), annotationClass);
        }

        // only found one, continue with that
        return fields.iterator().next();
    }

    /**
     * Provides a quick way to get a random, unused port by opening a {@link ServerSocket} and
     * getting the locally assigned port for the server socket.
     * 
     * @return A random, unused port.
     * 
     * @throws JitrException
     *             Thrown if any {@link IOException}s are thrown by the underlying calls to
     *             {@link ServerSocket}.
     */
    public static int getRandomUnusedPort() throws JitrException {

        final int port;
        ServerSocket socket = null;
        try {
            socket = new ServerSocket(0);
            port = socket.getLocalPort();
        } catch (final Exception e) {
            throw new JitrException("Failure picking a random, unused port.", e);
        } finally {
            if (socket != null) {
                try {
                    socket.close();
                } catch (final IOException ioe) {
                    throw new JitrException("Failure closing temporary socket.", ioe);
                }
            }
        }

        return port;
    }

    private static void validateClassesNotNull(final Class<?> clazz,
            final Class<? extends Annotation> annotationClass) {

        Validate.notNull(clazz, "Class is required.");
        Validate.notNull(annotationClass, "Annotation class is required.");
    }
}
